from typing import Dict, Iterable, Optional, Tuple, Union

import numpy as np
from cpsplines.psplines.bspline_basis import BsplineBasis


def granulate_prediction_range(
    bspline_bases: Iterable[BsplineBasis],
    prediction_step: Optional[Iterable[Iterable[Union[int, float]]]] = None,
) -> Tuple[Dict[int, np.ndarray]]:

    """
    Generate an array of equidistant points with small difference between each
    other at every prediction region.

    Parameters
    ----------
    bspline_bases : Iterable[BsplineBasis]
        The B-spline bases objects.
    prediction_step : Optional[Iterable[Iterable[Union[int, float]]]], optional
        The step length used to generate the granular regressor vector. The
        first term of each tuple corresponds to backwards region while the last
        term corresponds to forecast prediction. If None, a default step of 0.5
        on every direction is generated. By defualt, None.

    Returns
    -------
    Tuple[Dict[int, np.ndarray]]
        A tuple containing two dictionaries, one for backwards regions and one
        for forecast regions. Both dictionaries have as keys the ordinal names
        of the variables and as values the granular regressor vector on this
        region and for this variable.

    Raises
    ------
    ValueError
        If the number of B-spline bases and the prediction step passed have not
        the same shape.
    """

    if prediction_step is None:
        prediction_step = [(0.5, 0.5) for _ in range(len(bspline_bases))]
    if len({len(i) for i in [bspline_bases, prediction_step]}) != 1:
        raise ValueError(
            "The lengths of `bspline_bases` and `prediction_step` must agree."
        )

    x_left_pred = {}
    x_right_pred = {}
    for i, bsp in enumerate(bspline_bases):
        if bsp.int_back > 0:
            # This way, the minimum value of xsample is not picked up in the
            # np.arange of the backwards region
            x_left_pred[i] = np.arange(
                bsp.prediction["backwards"],
                bsp.xsample[0],
                prediction_step[i][0],
            )
        else:
            x_left_pred[i] = np.array([])

        if bsp.int_forw > 0:
            # This way, the maximum value of xsample is not picked up in the
            # np.arange of the forecast region
            x_right_pred[i] = np.arange(
                bsp.prediction["forward"],
                bsp.xsample[-1],
                -prediction_step[i][1],
            )[::-1]
        else:
            x_right_pred[i] = np.array([])
    return x_left_pred, x_right_pred



import numpy as np
from typing import Callable, Iterable, Union


class Simulator:

    """
    A callback function to display the current value of the arguments and the
    value of the objective function at each iteration. Based on the class
    developed in https://stackoverflow.com/a/59330005/4983192

    Parameters
    ----------
    func: Callable
        The objective function to be optimized.

    Attributes
    ----------
    callback_count: int
        Number of times the callback is called.
    sol_eval: list
        A list that contains all the intermediate solutions used in the problem.
    func_eval: list
        A list that contains all the evaluations of the objective function of the problem.
    """

    def __init__(self, function: Callable):
        self.func = function
        self.callback_count = 0
        self.sol_eval = []
        self.func_eval = []

    def simulate(self, x_k: Iterable[Union[int, float]], *args) -> Union[int, float]:

        """
        Executes the actual simulation and returns the result, while updating
        the attributes `sol_eval` and `func_eval`. This must be passed to the
        optimizer without arguments or parentheses.

        Inputs
        ------
        x_k: Iterable[Union[int, float]]
            The actual value for the solution.

        Returns
        -------
        (Union[int, float]) The objective function evaluated at x_k.
        """

        result = self.func(x_k, *args)
        self.sol_eval.append(x_k)
        self.func_eval.append(result)
        return result

    def callback(self, x_k: Iterable[Union[int, float]], *_) -> None:
        """
        Callback function that can be used by optimizers of scipy.optimize.
        The third argument "*_" makes sure that it still works when the
        optimizer calls the callback function with more than one argument. Pass
        to optimizer without arguments or parentheses.

        Inputs
        ------
        x_k: list
            The actual value for the solution.
        """

        # Locate the position in sol_eval that coincides with x_k. Once this
        # position is found, break the loop to save the position.
        for i, x in reversed(list(enumerate(self.sol_eval))):
            if np.allclose(x, x_k):
                break

        sp_info = ""
        # For each component in the actual solution, generate an string containing its
        # value and finally the value of the objective function stored in func_eval
        for comp in x_k:
            sp_info += f"{comp:10.5e}\t"
        sp_info += f"{self.func_eval[i]:10.5e}"

        # Set the title of the callback, with the smoothing parameter names and the
        # objective function label
        if not self.callback_count:
            title_list = [f"sp{j+1}" for j, _ in enumerate(x_k)] + ["Objective"]
            print("Starting the optimization algorithm")
            print(*title_list, sep="\t\t")
        # Print the actual solution and the actual objective function value
        print(sp_info)
        self.callback_count += 1
        return None



import numpy as np
from scipy.linalg import block_diag
import tensorly as tl
from typing import Iterable, List


def matrix_by_transpose(A: np.ndarray) -> np.ndarray:

    """
    Compute the product of a matrix by its transpose.

    Parameters
    ----------
    A : np.ndarray
        The input matrix

    Returns
    -------
    np.ndarray
        The product A.T @ A
    """

    return A.T @ A


def kronecker_matrix_by_identity(A: np.ndarray, n: int) -> np.ndarray:

    """
    Given a p x q matrix A and the identity matrix I_n, computes the
    Kronecker product np.kron(A, I_n). To compare performances with NumPy
    method, check https://stackoverflow.com/a/44461842/4983192

    Parameters
    ----------
    A : np.ndarray with shape (p, q)
        The input matrix to multiply.
    n : int
        The order of the identity matrix I_n.

    Returns
    -------
    np.ndarray with shape (np, nq)
        The Kronecker product np.kron(A, np.eye(n)).

    Raises
    ------
    ValueError
        If input array is not a matrix.
    """

    if A.ndim != 2:
        raise ValueError("Only two-dimensional arrays are allowed.")
    p, q = A.shape
    # The output is initialized as a 4th order array
    kron_prod = np.zeros((p, n, q, n))
    # Each of the values of A is repeated n times
    m = np.arange(n)
    # Index into the second and fourth axes and selecting all elements along
    # the rest to assign values from A. The values are broadcasted.
    kron_prod[:, m, :, m] = A
    # Finally reshape back to 2D
    return kron_prod.reshape(p * n, q * n)


def fast_kronecker_product(A: np.ndarray, B: np.ndarray) -> np.ndarray:

    """
    Given a m x n matrix A and a p x q matrix B, computes the Kronecker
    product np.kron(A, B) using the broadcasting operation. To compare
    performances with NumPy method, check
    https://stackoverflow.com/a/56067827/4983192

    Parameters
    ----------
    A : np.ndarray with shape (m, n)
        The m x n matrix A.
    B : np.ndarray with shape (p, q)
        The p x q matrix B.

    Returns
    -------
    np.ndarray with shape (mp, nq)
        The Kronecker product np.kron(A, B).

    Raises
    ------
    ValueError
        If any input array is not a matrix.
    """

    if A.ndim != 2 or B.ndim != 2:
        raise ValueError("Only two-dimensional arrays are allowed.")
    m, n = A.shape
    p, q = B.shape
    return (A[:, None, :, None] * B[None, :, None, :]).reshape(m * p, n * q)


def matrix_by_tensor_product(
    matrices: Iterable[np.ndarray], T: np.ndarray
) -> np.ndarray:

    """
    Given a n_1 x n_2 x ... x n_N multidimensional array T, computes efficiently
    the product T x_N A_N x_{N-1} ... x_1 A_1, where A_i are the elements of
    `matrices` and x_i represent the i-mode product.

    Parameters
    ----------
    matrices : Iterable[np.ndarray]
        A list containing matrices with dimensions m_i x n_i
    T : np.ndarray
        The n_1 x n_2 x ... x n_N multidimensional array.

    Returns
    -------
    np.ndarray
        The resulting m_1 x m_2 x ... x m_N multidimensional array.

    Raises
    ------
    ValueError
        If any input array is not a matrix.
    """
    if any(x.ndim != 2 for x in matrices):
        raise ValueError("Only two-dimensional arrays are allowed.")

    # Initialize computing the multidimensional array shape
    dims = list(T.shape)
    for i, A in enumerate(matrices):
        # Substitute the new dimension of the i-mode.
        dims[i] = A.shape[0]
        # The process consists of unfolding the multidimensional array
        # along the i-mode, then perform the product by the matrix of
        # the list and finally reshaping the matrix with the new shape
        T = tl.fold(A @ tl.unfold(T, i), i, dims)
    return T


def penalization_term(matrices: Iterable[np.ndarray]) -> List[np.ndarray]:

    """
    Given the penalty matrices defined over every axis, computes the
    penalization term of the objective function.

    Parameters
    ----------
    matrices : Iterable[np.ndarray]
        A list containing matrices with order n_i.

    Returns
    -------
    np.ndarray
        The penalization term items of the objective function, whose orders are
        the product over all the n_i.

    Raises
    ------
    ValueError
        If any input array is not a matrix.
    """

    if any(x.ndim != 2 for x in matrices):
        raise ValueError("Only two-dimensional arrays are allowed.")

    # Compute the order of the penalty matrices
    shapes = [P.shape[0] for P in matrices]
    # Initialize the output as a zero matrix
    output = []
    for i, P in enumerate(matrices):
        # Compute the shapes of the identity matrices located before
        # and after the penalty matrix for each summand
        left_id_shape = int(np.prod([s for s in shapes[:i]]))
        right_id_shape = int(np.prod([s for s in shapes[i + 1 :]]))
        # Compute the product np.kron(D, np.eye(right_identity_shape))
        right_kron_prod = kronecker_matrix_by_identity(P, right_id_shape)
        # Compute the product np.kron(np.eye(left_identity_shape), right_kron_prod)
        output.append(block_diag(*([right_kron_prod] * left_id_shape)))
    return output



import itertools
from functools import reduce
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import mosek.fusion
import numpy as np
import scipy
from cpsplines.mosek_functions.interval_constraints import IntConstraints
from cpsplines.mosek_functions.obj_function import ObjectiveFunction
from cpsplines.mosek_functions.point_constraints import PointConstraints
from cpsplines.psplines.bspline_basis import BsplineBasis
from cpsplines.psplines.penalty_matrix import PenaltyMatrix
from cpsplines.utils.cholesky_semidefinite import cholesky_semidef
from cpsplines.utils.fast_kron import (
    fast_kronecker_product,
    matrix_by_tensor_product,
    matrix_by_transpose,
    penalization_term,
)
from cpsplines.utils.gcv import GCV, gcv_mat
from cpsplines.utils.simulator_grid_search import print_grid_search_results
from cpsplines.utils.simulator_optimize import Simulator
from cpsplines.utils.weighted_b import get_idx_fitting_region, get_weighted_B
from joblib import Parallel, delayed


class NumericalError(Exception):
    pass


class GridCPsplines:

    """
    Create the constrained P-splines model when data are in array form. The
    objective function of the optimization problem is defined using the approach
    of Currie et al. (2006) and the constraints are imposed over the whole
    domain with univariate data and over a finite set of curves (coinciding with
    the position of the knot sequences) otherwise.

    Parameters
    ----------
    deg : Iterable[int], optional
        The polynomial degree vector of the B-spline bases. All the elements
        must be non-negative integers. By default, (3,).
    ord_d : Iterable[int], optional
        The difference order vector of the penalty matrices. All the elements
        must be non-negative integers. By default, (2,).
    n_int : int, optional
        The vector of number of equal intervals which the fitting region along
        each dimension is split. All the elements must be non-negative integers.
        By default, (40,).
    x_range : Optional[Dict[int, Tuple[Union[int, float]]]], optional
        A dictionary containing the most extreme values that the extended
        B-spline basis needs to capture along each dimension. The keys are index
        of the regressor and the values are tuples containing the most extreme
        values. These values will only be considered when they are located
        outside the fitting region. One or two may be passed by variable. If
        None, it is intended that no extra range is needed in any of the axes.
        By default, None.
    sp_method : str, optional
        The method used to estimate the smoothing parameter as the minimizer of
        the Generalized Cross Validation criterium. It may be either `optimizer`
        (to use minimization algorithms in scipy.optimize.minimize) or
        `grid_search` (to select the smoothing parameters over a grid of
        candidates). By default, "optimizer".
    sp_args : Optional[Dict[str, Any]], optional
        The arguments employed on the chosen `sp_method`.
        1. If `grid_search` is selected, they are:
        - "grid" (Iterable[Iterable[Union[int, float]]]) : The smoothing
        parameter candidates along each axis. By default, (0.01, 0.1, 1, 10) on
        each direction.
        - "parallel" (bool) : If True, the different combinations of smoothing
        parameters are run in parallel using `joblib` module. By default, False.
        - "n_jobs" (int) : The number of jobs to run in parallel. By default,
        -2.
        - "top_n" (Optional[int]): If not None, prints the best `n` combinations
        of smoothing parameters.
        2. If `optimizer` is selected, they are:
        - "verbose" (bool) : If True, a callback printing the solution and the
        values at each iteration is used. By default, False.
        - The optional arguments and the first guess from
        scipy.optimize.minimize can be passed on this dictionary. By default,
        the SLSQP solver with a vector of ones as first guess is used. Also, the
        smoothing parameters are constrained to be in (1e-10, 1e16) since they
        must be non-negative.
    int_constraints : Dict[int, Dict[int, Dict[str, Union[int, float]]]]],
    optional
        A nested dictionary containing the interval constraints to be enforced.
        The keys of the dictionary are the indexes of the variables. The values
        are dictionaries, whose keys are the order of the derivative. The values
        are dictionareis, whose keys are the signs of the constraints (either
        "+" or "-") and the values are numbers denoting the upper or lower
        threshold of the constraints. By default, None.
    pt_constraints : Optional[Dict[Tuple[int], Any]], optional
        A dictionary containing the point constraints to be enforced. The keys
        of the dictionary are tuples representing the order of the derivatives
        where the constraints acts on. The values are tuples that must contain
        the items (in this order):
        - An array of unidimensional arrays with the coordinates of the points
        where the value needs to be fixed.
        - An array with the values of the derivative to be enforced.
        - A number corresponding to the tolerancea allowed in the constraint.

    Attributes
    ----------
    bspline_bases : List[BsplineBasis]
        The list of B-spline bases on each axis resulting from
        `_get_bspline_bases`.
    sol : np.ndarray
        The fitted decision variables of the B-spline expansion.
    y_fitted : np.ndarray
        The fitted values for the response variable.

    References
    ----------
    - Currie, I. D., Durban, M., & Eilers, P. H. (2006). Generalized linear
      array models with applications to multidimensional smoothing. Journal of
      the Royal Statistical Society: Series B (Statistical Methodology), 68(2),
      259-280.
    """

    def __init__(
        self,
        deg: Iterable[int] = (3,),
        ord_d: Iterable[int] = (2,),
        n_int: Iterable[int] = (40,),
        x_range: Optional[Dict[int, Tuple[Union[int, float]]]] = None,
        sp_method: str = "optimizer",
        sp_args: Optional[Dict[str, Any]] = None,
        int_constraints: Optional[
            Dict[int, Dict[int, Dict[str, Union[int, float]]]]
        ] = None,
        pt_constraints: Optional[Dict[Tuple[int], Any]] = None,
    ):
        self.deg = deg
        self.ord_d = ord_d
        self.n_int = n_int
        self.x_range = x_range
        self.sp_method = sp_method
        self.sp_args = sp_args
        self.int_constraints = int_constraints
        self.pt_constraints = pt_constraints

    def _get_bspline_bases(self, x: Iterable[np.ndarray]) -> List[BsplineBasis]:

        """
        Construct the B-spline bases on each axis.

        Parameters
        ----------
        x : Iterable[np.ndarray]
            The covariates samples.

        Returns
        -------
        List[BsplineBasis]
            The list of B-spline bases on each axis.
        """

        bspline_bases = []
        if self.x_range is None:
            self.x_range = {}
        for i in range(len(self.deg)):
            # Get the maximum and minimum of the fitting regions
            x_min, x_max = np.min(x[i]), np.max(x[i])
            prediction_dict = {}
            if i in self.x_range.keys():
                # If the values in `x_range` are outside the fitting region,
                # include them in the `prediction` argument of the BsplineBasis
                pred_min, pred_max = min(self.x_range[i]), max(self.x_range[i])
                if pred_max > x_max:
                    prediction_dict["forward"] = pred_max
                if pred_min < x_min:
                    prediction_dict["backwards"] = pred_min
            bsp = BsplineBasis(
                deg=self.deg[i],
                xsample=x[i],
                n_int=self.n_int[i],
                prediction=prediction_dict,
            )
            # Generate the design matrix of the B-spline basis
            bsp.get_matrix_B()
            bspline_bases.append(bsp)
        return bspline_bases

    def _fill_sp_args(self):

        """
        Fill the `sp_args` dictionary by default parameters on the case they are
        not provided.
        """

        if self.sp_args is None:
            self.sp_args = {}
        if self.sp_method == "grid_search":
            self.sp_args["grid"] = self.sp_args.get(
                "grid", [(0.01, 0.1, 1, 10) for _ in range(len(self.deg))]
            )
            self.sp_args["parallel"] = self.sp_args.get("parallel", False)
            self.sp_args["n_jobs"] = self.sp_args.get("n_jobs", -2)
            self.sp_args["top_n"] = self.sp_args.get("top_n", None)
        else:
            self.sp_args["verbose"] = self.sp_args.get("verbose", False)
            self.sp_args["x0"] = self.sp_args.get("x0", np.ones(len(self.deg)))
            self.sp_args["method"] = self.sp_args.get("method", "SLSQP")
            self.sp_args["bounds"] = self.sp_args.get(
                "bounds", [(1e-10, 1e16) for _ in range(len(self.deg))]
            )
        return None

    def _get_obj_func_arrays(self, y: np.ndarray) -> Dict[str, np.ndarray]:

        """
        Gather all the arrays used to define the objective function of the
        optimization funcion. These are the design matrices of the B-spline
        basis, the penalty matrices and the extended response variable sample
        (which is equal to zero outside the fitting region and coincides with
        the response variable sample inside it).

        Parameters
        ----------
        y : np.ndarray
            The response variable sample.

        Returns
        -------
        Dict[str, np.ndarray]
            A dictionary containing design matrices of the B-spline basis, the
            penalty matrices and the extended response variable sample.
        """

        obj_matrices = {}
        obj_matrices["B"] = []
        obj_matrices["D_mul"] = []
        # The extended response variable sample dimensions can be obtained as
        # the number of rows of the design matrix B
        y_ext_dim = []
        for i, bsp in enumerate(self.bspline_bases):
            B = bsp.matrixB
            y_ext_dim.append(B.shape[0])
            obj_matrices["B"].append(B)
            obj_matrices["D_mul"].append(
                PenaltyMatrix(bspline=bsp).get_penalty_matrix(
                    **{"ord_d": self.ord_d[i]}
                )
            )

        y_ext = np.zeros(tuple(y_ext_dim))
        y_ext[get_idx_fitting_region(self.bspline_bases)] = y
        obj_matrices["y"] = y_ext
        return obj_matrices

    def _initialize_model(
        self, lin_term: np.ndarray, L_B: np.ndarray, L_D: Iterable[np.ndarray]
    ) -> mosek.fusion.Model:

        """
        Construct the optimization model.

        Parameters
        ----------
        lin_term : np.ndarray
            An array containing the coefficients of the linear term.
        L_B : np.ndarray
            The Cholesky decomposition of B.T @ B, where B is the Kronecker
            product of the B-spline basis matrices.
        L_D : Iterable[np.ndarray]
            An array containing the Cholesky decomposition of P_i, where P_i is
            the penalty matrix along the i axis.

        Returns
        -------
        mosek.fusion.Model
            The optimization model.
        """

        M = mosek.fusion.Model()
        # Create the variables of the optimization problem
        mos_obj_f = ObjectiveFunction(bspline=self.bspline_bases, model=M)
        # For each axis, a smoothing parameter is needed
        sp = [M.parameter(f"sp_{i}", 1) for i in range(len(L_D))]
        # Build the objective function of the problem
        mos_obj_f.create_obj_function(
            L_B=L_B,
            L_D=L_D,
            sp=sp,
            lin_term=lin_term,
        )

        if self.int_constraints is not None:
            matrices_S = {}
            # If the are interval constraints, construct the matrices S
            for i, bsp in enumerate(self.bspline_bases):
                matrices_S[i] = bsp.get_matrices_S()
            # Iterate for every variable with constraints and for every
            # derivative order
            for var_name in self.int_constraints.keys():
                for deriv in self.int_constraints[var_name].keys():
                    matrices_S_copy = matrices_S.copy()
                    # Build the interval constraints
                    cons = IntConstraints(
                        bspline=self.bspline_bases,
                        var_name=var_name,
                        derivative=deriv,
                        constraints=self.int_constraints[var_name][deriv],
                    )
                    cons.interval_cons(
                        var_dict=mos_obj_f.var_dict, model=M, matrices_S=matrices_S_copy
                    )
        else:
            self.int_constraints = {}

        if self.pt_constraints is not None:
            # Iterate for every combination of the derivative orders where
            # constraints must be enforced
            for deriv, info in self.pt_constraints.items():
                cons2 = PointConstraints(
                    pts=info[0],
                    value=info[1],
                    derivative=deriv,
                    bspline=self.bspline_bases,
                    tolerance=info[2],
                )
                cons2.point_cons(var_dict=mos_obj_f.var_dict, model=M)
        else:
            self.pt_constraints = {}

        return M

    def _get_sp_grid_search(
        self,
        B_weighted: Iterable[np.ndarray],
        Q_matrices: Iterable[np.ndarray],
        y: np.ndarray,
    ) -> Tuple[Union[int, float]]:

        """
        Get the best smoothing parameter vector with the GCV minimizer criteria
        using grid search selection.

        Parameters
        ----------
        B_weighted : Iterable[np.ndarray]
            The weighted design matrix from the B-spline basis.
        Q_matrices : Iterable[np.ndarray]
            The array of matrices used in the GCV computation.
        y : np.ndarray
            The extended response variable sample.

        Returns
        -------
        Tuple[Union[int, float]]
            The best set of smoothing parameters
        """

        # Computes all the possible combinations for the smoothing parameters
        iter_sp = list(itertools.product(*self.sp_args["grid"]))
        # Run in parallel if the argument `parallel` is present
        if self.sp_args["parallel"] == True:
            gcv = Parallel(n_jobs=self.sp_args["n_jobs"])(
                delayed(GCV)(sp, B_weighted, Q_matrices, y) for sp in iter_sp
            )
        else:
            gcv = [
                GCV(
                    sp=sp,
                    B_weighted=B_weighted,
                    Q_matrices=Q_matrices,
                    y=y,
                )
                for sp in iter_sp
            ]
        # Print the `top_n` combinations that minimizes the GCV
        if self.sp_args["top_n"] is not None:
            print_grid_search_results(
                x_val=iter_sp, obj_val=gcv, top_n=self.sp_args["top_n"]
            )

        return iter_sp[gcv.index(min(gcv))]

    def _get_sp_optimizer(
        self,
        B_weighted: Iterable[np.ndarray],
        Q_matrices: Iterable[np.ndarray],
        y: np.ndarray,
    ) -> Tuple[Union[int, float]]:

        """
        Get the best smoothing parameter vector with the GCV minimizer criteria
        using an optimizer from scipy.optimize.minimize.

        Parameters
        ----------
        B_weighted : Iterable[np.ndarray]
            The weighted design matrix from the B-spline basis.
        Q_matrices : Iterable[np.ndarray]
            The array of matrices used in the GCV computation.
        y : np.ndarray
            The extended response variable sample.

        Returns
        -------
        Tuple[Union[int, float]]
            The best set of smoothing parameters
        """

        # All argument in `sp_args` except `verbose` are arguments from
        # scipy.optimize.minimize, so create a copy without it
        scipy_optimize_params = self.sp_args.copy()
        scipy_optimize_params.pop("verbose", None)
        # Create a simulator to print the intermediate steps of the process if
        # "verbose" is active
        if self.sp_args["verbose"]:
            gcv_sim = Simulator(GCV)
        # Get the best set of smoothing parameters
        best_sp = scipy.optimize.minimize(
            gcv_sim.simulate if self.sp_args["verbose"] else GCV,
            args=(
                B_weighted,
                Q_matrices,
                y,
            ),
            callback=gcv_sim.callback if self.sp_args["verbose"] else None,
            **scipy_optimize_params,
        ).x
        return best_sp

    def fit(self, x: Iterable[np.ndarray], y: np.ndarray):

        """
        Compute the fitted decision variables of the B-spline expansion and the
        fitted values for the response variable.

        Parameters
        ----------
        x : Iterable[np.ndarray]
            The covariate samples.
        y : np.ndarray
            The response variable sample.

        Raises
        ------
        ValueError
            If the degree, difference order, number of interval and coordinates
            vectors length differs.
        ValueError
            If `sp_method` input is different from "grid_search" or "optimizer".
        NumericalError
            If MOSEK could not arrive to a feasible solution.
        """

        if len({len(i) for i in [self.deg, self.ord_d, self.n_int, x]}) != 1:
            raise ValueError(
                "The lengths of `deg`, `ord_d`, `n_int` and `x` must agree."
            )

        if self.sp_method not in ["grid_search", "optimizer"]:
            raise ValueError(f"invalid `sp_method`: {self.sp_method}")

        # Construct the B-spline bases
        self.bspline_bases = self._get_bspline_bases(x=x)

        # Filling the arguments of the method used to determine the optimal set
        # of smoothing parameters
        _ = self._fill_sp_args()

        # Get the matrices used in the objective function
        obj_matrices = self._get_obj_func_arrays(y=y)

        # Auxiliary matrices derived from `obj_matrices`
        B_weighted = get_weighted_B(bspline_bases=self.bspline_bases)
        obj_matrices["B_mul"] = list(map(matrix_by_transpose, B_weighted))
        # Compute the linear term coefficients of the objective function
        lin_term = np.multiply(
            -2,
            matrix_by_tensor_product([mat.T for mat in B_weighted], obj_matrices["y"]),
        ).flatten()

        # Compute the Cholesky decompositions (A = L @ L.T)
        L_B = reduce(
            fast_kronecker_product, list(map(cholesky_semidef, obj_matrices["B_mul"]))
        )

        L_D = penalization_term(
            matrices=list(map(cholesky_semidef, obj_matrices["D_mul"]))
        )

        # Initialize the model
        M = self._initialize_model(lin_term=lin_term, L_B=L_B, L_D=L_D)
        model_params = {"theta": M.getVariable("theta")}
        for i in range(len(self.bspline_bases)):
            model_params[f"sp_{i}"] = M.getParameter(f"sp_{i}")

        # Get the matrices used in the GCV computation
        Q_matrices = gcv_mat(B_mul=obj_matrices["B_mul"], D_mul=obj_matrices["D_mul"])

        if self.sp_method == "grid_search":
            self.best_sp = self._get_sp_grid_search(
                B_weighted=B_weighted, Q_matrices=Q_matrices, y=obj_matrices["y"]
            )
        else:
            self.best_sp = self._get_sp_optimizer(
                B_weighted=B_weighted, Q_matrices=Q_matrices, y=obj_matrices["y"]
            )
        theta_shape = model_params["theta"].getShape()
        # Set the smoothing parameters vector as the optimal obtained in the
        # unconstrained setting
        for i, sp in enumerate(self.best_sp):
            model_params[f"sp_{i}"].setValue(sp)
        try:
            # Solve the problem
            M.solve()
            # Extract the fitted decision variables of the B-spline expansion
            self.sol = model_params["theta"].level().reshape(theta_shape)
            # Compute the fitted values of the response variable
            self.y_fitted = matrix_by_tensor_product(
                [mat for mat in obj_matrices["B"]], self.sol
            )
        except mosek.fusion.SolutionError as e:
            raise NumericalError(
                f"The solution for the smoothing parameter {self.best_sp} "
                f"could not be found due to numerical issues. The original error "
                f"was: {e}"
            )

        return None

    def predict(self, x: Iterable[np.ndarray]) -> np.ndarray:
        # Return empty dataset when not all the coordinates required are passed
        if len([v for v in x if len(v) > 0]) < len(x):
            return np.array([])
        else:
            x_min = np.array([np.min(v) for v in x])
            x_max = np.array([np.max(v) for v in x])
            bsp_min = np.array([bsp.knots[bsp.deg] for bsp in self.bspline_bases])
            bsp_max = np.array([bsp.knots[-bsp.deg] for bsp in self.bspline_bases])
            # If some coordinates are outside the range where the B-spline bases
            # were defined, the problem must be fitted again
            if (x_min < bsp_min).sum() > 0:
                raise ValueError(
                    f"Some of the coordinates are outside the definition range of "
                    f"the B-spline bases."
                )
            if (x_max > bsp_max).sum() > 0:
                raise ValueError(
                    f"Some of the coordinates are outside the definition range of "
                    f"the B-spline bases."
                )
            # Compute the basis matrix at the coordinates to be predicted
            B_predict = [
                bsp.bspline_basis.derivative(nu=0)(x[i])
                for i, bsp in enumerate(self.bspline_bases)
            ]
            # Get the predictions
            return matrix_by_tensor_product([mat for mat in B_predict], self.sol)



import numpy as np
import pandas as pd
from typing import Union


def normalize_data(x: Union[np.ndarray, pd.Series]) -> Union[np.ndarray, pd.Series]:

    """
    Normalize an array between 0 and 1.

    Parameters
    ----------
    x : Union[np.ndarray, pd.Series]
        The data to be normalized.

    Returns
    -------
    Union[np.ndarray, pd.Series]
        The data normalized.
    """

    return (x - np.min(x)) / np.ptp(x)



from functools import reduce

import numpy as np
import pytest
from cpsplines.utils.fast_kron import matrix_by_tensor_product
from cpsplines.utils.timer import timer


# Test the efficiency of folding and unfolding matrices compared to brute force
# approximation
@pytest.mark.parametrize(
    "common_dim, different_dim",
    [
        ([200], [23]),
        ([150, 175], [43, 67]),
        ([75, 53, 30], [7, 8, 9]),
    ],
)
def test_matrices_by_tensor(common_dim, different_dim):
    random_tensor = np.random.rand(*common_dim)
    random_mat = [np.random.rand(c, d) for c, d in zip(common_dim, different_dim)]

    with timer(tag="Using np.kron"):
        exp_out = (random_tensor.flatten() @ reduce(np.kron, random_mat)).reshape(
            *different_dim
        )

    with timer(tag="Using folding and unfolding"):
        out = matrix_by_tensor_product(
            matrices=[A.T for A in random_mat], T=random_tensor
        )

    np.testing.assert_allclose(exp_out, out)



from typing import Iterable, List, Tuple

import numpy as np
from cpsplines.psplines.bspline_basis import BsplineBasis


def get_idx_fitting_region(bspline_bases: Iterable[BsplineBasis]) -> Tuple[slice]:

    """
    Get the fitting region indices on the expanded sample regressor vector
    (containing the extra knots from the extended B-spline basis and the sample
    regressor vector).

    Parameters
    ----------
    bspline_bases : Iterable[BsplineBasis]
        The B-spline bases.

    Returns
    -------
    Tuple[slice]
        Tuple of slices ranging from the first to the last index in the fitting
        region, one by axis.
    """

    return tuple(
        slice(bsp.int_back, bsp.int_back + len(bsp.xsample), None)
        for bsp in bspline_bases
    )


def get_weighted_B(bspline_bases: Iterable[BsplineBasis]) -> List[np.ndarray]:

    """
    Get the weighted design matrices from the B-spline bases, which assing a
    zero weight to the rows of the values to be predicted and one weight to the
    rows on the fitting region.

    Parameters
    ----------
    bspline_bases : Iterable[BsplineBasis]
        The B-spline bases.

    Returns
    -------
    List[np.ndarray]
        The weighted matrices B.
    """

    mat = []
    idx_fit = get_idx_fitting_region(bspline_bases=bspline_bases)
    for i, bsp in enumerate(bspline_bases):
        B_weighted = np.zeros((bsp.matrixB.shape))
        B_weighted[idx_fit[i], :] = bsp.matrixB[idx_fit[i], :]
        mat.append(B_weighted)
    return mat



import matplotlib.pyplot as plt
import matplotlib
import numpy as np
import random
from typing import Iterable, Optional, Tuple, Union

from cpsplines.fittings.grid_cpsplines import GridCPsplines
from cpsplines.graphics.plot_utils import granulate_prediction_range


def plot_curves(
    fittings: Iterable[GridCPsplines],
    col_curve: Optional[Iterable[str]] = None,
    x: Optional[Iterable[np.ndarray]] = None,
    y: Optional[Iterable[np.ndarray]] = None,
    col_pt: Optional[Iterable[str]] = None,
    alpha: Union[int, float] = 0.25,
    prediction_step: Iterable[Union[int, float]] = (0.5, 0.5),
    knot_positions: bool = False,
    constant_constraints: bool = False,
    figsize: Tuple[Union[int, float]] = (15, 10),
) -> Tuple[matplotlib.figure.Figure, plt.axes]:

    """
    Plot a set of curves fitted using the method `GridCPsplines`.

    Parameters
    ----------
    fittings : Iterable[GridCPsplines]
        An iterable of fitted `GridCPsplines` objects
    col_curve : Optional[Iterable[str]], optional
        An iterable with the colours used to graph the curves (in the same order
        as the corresponding curve). If None, the colours are chosen as random.
        By default, None.
    x : Optional[Iterable[np.ndarray]], optional
        An iterable with the x-coordinates of points to be plotted. Each element
        contain a set of coordinates. By default, None.
    y : Optional[Iterable[np.ndarray]], optional
        An iterable with the y-coordinates of points to be plotted. By default,
        None.
    col_pt : Optional[Iterable[str]], optional
        An iterable with the colours used to plot the points (in the same order
        as the corresponding set of points). If None, the colours are chosen as
        random. By default, None.
    alpha : Union[int, float], optional
        The transparency level of the points. Only applies when points are
        plotted. By default, 0.25.
    prediction_step : Iterable[Union[int, float]], optional
        The step used to produce equidistant extra points at the prediction
        regions so the graph of the curves seems smoother (if this is not
        applied, the only points at the prediction points are the knots). The
        first element of the iterable is the step on the backwards prediction
        and the second element corresponds to the step on the forward
        prediction. By default, (0.5, 0.5).
    knot_positions : bool, optional
        If True, the positions where the inner knots are located are marked as
        grey vertical lines. By default, False.
    constant_constraints : bool, optional
        If True, horizontal lines at the threshold of the zero-order derivative
        constraints are plotted with red dashed lines. By default, False.
    figsize : Tuple[Union[int, float]], optional
        The size of the figure. By default, (15, 10).

    Returns
    -------
    Tuple[matplotlib.figure.Figure, plt.axes]
        A tuple containing the figure object and the axis.

    Raises
    ------
    ValueError
        If the number of the curve colours do not coincide with the number of
        curves.
    ValueError
        If the number of the points colours do not coincide with the number of
        set of points.
    AttributeError
        If any of the curve objects is not fitted yet.
    """

    fig, ax = plt.subplots(figsize=figsize)
    # If `col_curve` is None, create the colours randomly
    if col_curve is None:
        col_curve = [
            "#" + "".join([random.choice("0123456789ABCDEF") for _ in range(6)])
            for _ in range(len(fittings))
        ]

    if len(col_curve) != len(fittings):
        raise ValueError("Number of colours and number of curves must agree.")

    # If there are set of points to plot, get a random colour if `col_pt` is
    # None and get the scatter plot.
    if (x is not None) and (y is not None):
        if col_pt is None:
            col_pt = [
                "#" + "".join([random.choice("0123456789ABCDEF") for _ in range(6)])
                for _ in range(len(x))
            ]
        if len({len(i) for i in [x, y, col_pt]}) != 1:
            raise ValueError("The lengths of `x`, `y` and `col_pt` must agree.")
        for i, (pt_x, pt_y) in enumerate(zip(x, y)):
            _ = ax.scatter(
                x=pt_x,
                y=pt_y,
                c=col_pt[i],
                alpha=alpha,
            )

    for i, curve in enumerate(fittings):
        # The curves need to be fitted
        if not hasattr(curve, "y_fitted"):
            raise AttributeError("All curves must be fitted.")
        bsp = curve.bspline_bases[0]
        # Generate extra points at the prediction regions with the
        # `prediction_step` parameter so the curves are plotted smoother
        x_left, x_right = granulate_prediction_range(
            bspline_bases=[bsp], prediction_step=[prediction_step]
        )
        # Define the number of points to be removed since they are in the
        # forecasting region
        y_end = None if bsp.int_forw == 0 else -bsp.int_forw

        # Plot the curves using the extended fitted coordinates
        _ = ax.plot(
            np.concatenate([x_left[0], bsp.xsample, x_right[0]]),
            np.concatenate(
                [
                    curve.predict([x_left[0]]),
                    curve.y_fitted[bsp.int_back : y_end],
                    curve.predict([x_right[0]]),
                ]
            ),
            col_curve[i],
            linewidth=2.0,
            label=str(i),
        )

        # If the prediction region is not empty, plot vertical dashed lines at
        # the extremes of the fitting region
        if bsp.int_back > 0:
            _ = ax.axvline(
                bsp.xsample[0], color=col_curve[i], linewidth=1.0, linestyle="--"
            )
        if bsp.int_forw > 0:
            _ = ax.axvline(
                bsp.xsample[-1], color=col_curve[i], linewidth=1.0, linestyle="--"
            )
        # If it is required, plot the position of the knots
        if knot_positions:
            for knot in bsp.knots[bsp.deg : -bsp.deg]:
                _ = ax.axvline(knot, color="grey", alpha=0.25)
        # If it is required, threshold of the zero-order derivative constraints
        if constant_constraints:
            if curve.int_constraints:
                if 0 in curve.int_constraints[0].keys():
                    for value in curve.int_constraints[0][0].values():
                        _ = ax.axhline(
                            value,
                            color="red",
                            linewidth=1.0,
                            linestyle="--",
                        )
    return fig, ax


def plot_surfaces(
    fittings: Iterable[GridCPsplines],
    col_surface: Optional[Iterable[str]] = None,
    contour_plot: bool = True,
    prediction_step: Iterable[Iterable[Union[int, float]]] = ((0.5, 0.5), (0.5, 0.5)),
    figsize: Tuple[Union[int, float]] = (12, 6),
    knot_positions: bool = False,
    zlim: Optional[Tuple[Union[int, float]]] = None,
    orientation: Optional[Tuple[Union[int, float]]] = None,
) -> Tuple[Tuple[matplotlib.figure.Figure, plt.axes]]:

    """
    Plot a set of curves fitted using the method `GridCPsplines`.

    Parameters
    ----------
    fittings : Iterable[GridCPsplines]
        An iterable of fitted `GridCPsplines` objects.
    col_surface : Optional[Iterable[str]], optional
        An iterable with the colours used to graph the surfaces (in the same
        order as the corresponding curve). If None, the colours are chosen as
        random. By default, None.
    contour_plot : bool, optional
        If True, the contour plot of the surface is plotted with the same
        colormap that the surface plot. By default, True.
    prediction_step : Iterable[Iterable[Union[int, float]]], optional
        The step used to produce equidistant extra points at the prediction
        regions so the graph of the surfaces seems smoother. The first element
        of the iterable corresponds to the first direction while the second
        element to the second direction. For each tuple, first element is the
        step on the backwards prediction and the second element corresponds to
        the step on the forward prediction. By default, ((0.5, 0.5), (0.5, 0.5)).
    figsize : Tuple[Union[int, float]], optional
        The size of the figure. By default, (12, 6).
    knot_positions : bool, optional
        If True, the positions where the inner knots are located are marked as
        red dashed lines. By default, False.
    zlim : Optional[Tuple[Union[int, float]]], optional
        An iterable with two elements used to restrict the range on the z-axis.
        First element is the lower bound and second element corresponds to upper
        bound. By default, None.
    orientation : Optional[Tuple[Union[int, float]]], optional
        Set the elevation angle in the z plane and azimuth angle in the xy plane
        of the axes. If None, the default value is (30,-60).

    Returns
    -------
    Tuple[Tuple[matplotlib.figure.Figure, plt.axes]]
        A tuple of tuples containing the figure objects and the axes.

    Raises
    ------
    ValueError
        If wrong values are passed to `orientation`.
    ValueError
        If wrong values are passed to `zlim`.
    ValueError
        If the number of the surface colours do not coincide with the number of
        surfaces.
    AttributeError
        If any of the surface objects is not fitted yet.
    """

    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111, projection="3d")
    # Provide the required orientation
    if orientation is not None:
        if len(orientation) != 2:
            raise ValueError("Only two angle coordinates may be passed.")
        _ = ax.view_init(*orientation)
    # Restrict to the required bounds the z-axis.
    if zlim is not None:
        if len(orientation) != 2:
            raise ValueError("Only two z-axis limits may be passed.")
        _ = ax.set_zlim3d(zlim)
        kwargs_surf = {"vmin": zlim[0], "vmax": zlim[1]}
    else:
        kwargs_surf = {}
    # If no colormaps are provided, get random colormaps
    if col_surface is None:
        col_surface = random.sample(plt.colormaps(), len(fittings))

    if len(col_surface) != len(fittings):
        raise ValueError("Number of colours and number of surfaces must agree.")
    fig_ax = [(fig, ax)]

    for i, surface in enumerate(fittings):
        if not hasattr(surface, "y_fitted"):
            raise AttributeError("All surfaces must be fitted.")

        # Generate extra points at the prediction regions with the
        # `prediction_step` parameter so the surfaces are plotted smoother
        x_left, x_right = granulate_prediction_range(
            bspline_bases=surface.bspline_bases, prediction_step=prediction_step
        )
        # Get the extended regressor samples
        ext_x0 = np.concatenate(
            [x_left[0], surface.bspline_bases[0].xsample, x_right[0]]
        )
        ext_x1 = np.concatenate(
            [x_left[1], surface.bspline_bases[1].xsample, x_right[1]]
        )
        # Define the number of points to be removed since they are in the
        # forecasting regions
        y_end_0 = (
            None
            if surface.bspline_bases[0].int_forw == 0
            else -surface.bspline_bases[0].int_forw
        )
        y_end_1 = (
            None
            if surface.bspline_bases[1].int_forw == 0
            else -surface.bspline_bases[1].int_forw
        )

        ext_y = surface.y_fitted[
            surface.bspline_bases[0].int_back : y_end_0,
            surface.bspline_bases[1].int_back : y_end_1,
        ]
        # Add the predictions of the new points to the extended response matrix
        if x_left[1].size > 0:
            pred_left = surface.predict(x=[surface.bspline_bases[0].xsample, x_left[1]])
            ext_y = np.concatenate([pred_left, ext_y], axis=1)
        if x_right[1].size > 0:
            pred_right = surface.predict(
                x=[surface.bspline_bases[0].xsample, x_right[1]]
            )
            ext_y = np.concatenate([ext_y, pred_right], axis=1)
        if x_left[0].size > 0:
            pred_up = surface.predict(x=[x_left[0], ext_x1])
            ext_y = np.concatenate([pred_up, ext_y], axis=0)
        if x_right[0].size > 0:
            pred_down = surface.predict(x=[x_right[0], ext_x1])
            ext_y = np.concatenate([ext_y, pred_down], axis=0)

        # Plot the surface and include the colorbar
        Z, X = np.meshgrid(ext_x0, ext_x1)
        surf = ax.plot_surface(
            Z, X, ext_y.T, cmap=col_surface[i], rstride=2, cstride=2, **kwargs_surf
        )
        _ = fig.colorbar(surf, ax=ax)

        # If required, plot the contour plot of the surface
        if contour_plot:
            fig_contour, ax_contour = plt.subplots(figsize=figsize)
            _ = ax_contour.contourf(
                Z, X, ext_y.T, 100, cmap=col_surface[i], **kwargs_surf
            )
            _ = fig_contour.colorbar(surf, ax=ax_contour)
            # Establish the limits on the two axis (otherwise some extra knots
            # in the prediction regions may be further apart from extreme points
            # of the extended regressor samples)
            _ = ax_contour.set_xlim(ext_x0[0], ext_x0[-1])
            _ = ax_contour.set_ylim(ext_x1[0], ext_x1[-1])
            fig_ax.append((fig_contour, ax_contour))
        # If it is required, plot the position of the knots
        if knot_positions:
            for knot in surface.bspline_bases[0].knots[
                surface.bspline_bases[0].deg : -surface.bspline_bases[0].deg
            ]:
                _ = ax_contour.plot(
                    [knot] * len(ext_x1),
                    ext_x1,
                    color="red",
                    linestyle="--",
                    alpha=0.3,
                )
            for knot in surface.bspline_bases[1].knots[
                surface.bspline_bases[1].deg : -surface.bspline_bases[1].deg
            ]:
                _ = ax_contour.plot(
                    ext_x0,
                    [knot] * len(ext_x0),
                    color="red",
                    linestyle="--",
                    alpha=0.3,
                )
    return tuple(fig_ax)



from contextlib import contextmanager
from timeit import default_timer
from typing import Optional


@contextmanager
def timer(tag: Optional[str] = None) -> None:

    """
    Computes the elapsed time that a task last.

    Parameters
    ----------
    tag : str, optional
        The name of the task. By default, None.
    """

    start = default_timer()
    try:
        yield
    finally:
        end = default_timer()
        header = "Elapsed time (s)" if tag is None else f"[{tag}] Elapsed time (s)"
        print(f"{header}: {end - start:.6f}")



from typing import Callable, Dict, Iterable, Optional, Union

import numpy as np
import pandas as pd
from cpsplines.utils.normalize_data import normalize_data


def filter_covid_df(
    response_var: str,
    df: Optional[pd.DataFrame] = None,
    min_date: Optional[pd.Timestamp] = None,
    max_date: Optional[pd.Timestamp] = None,
    gender: Optional[Iterable[str]] = None,
    region: Optional[Iterable[str]] = None,
) -> pd.DataFrame:

    """
    Filter the COVID-19 DataFrame generated by Instituto de Salud Carlos III
    (https://cnecovid.isciii.es/covid19/).

    Parameters
    ----------
    response_var : str
        The response variable to be chosen. It must be one of the following:
            - "num_casos": Number of infected people.
            - "num_hosp": Number of hospitalized patients.
            - "num_uci": Number of ICU hospitalized patients.
            - "num_def": Number of fatalities.
    df : Optional[pd.DataFrame], optional
        The COVID-19 DataFrame. If None, it is downloaded from the website. By
        default, None.
    min_date : Optional[pd.Timedelta], optional
        The minimum date to be considered. If None, it coincides with the first
        non-zero response variable date. By default, None.
    max_date : Optional[pd.Timedelta], optional
        The maximum date to be considered. If None, it coincides with the
        maximum date of the DataFrame. By default, None.
    gender : Optional[Iterable[str]], optional
        The gender (male and/or female) to be considered. If None, both genders
        and undetermined gender ("NC") are taken. By default, None.
    region : Optional[Iterable[str]], optional
        The province ISO code to be considered. The allowed codes are available
        on the province_es.csv on the column "code". If None, all regions  are
        taken. By default, None.

    Returns
    -------
    pd.DataFrame
        The filtered DataFrame.

    Raises
    ------
    ValueError
        If `response_var` is not an allowed response variable.
    ValueError
        If `gender` are not allowed gender categories.
    ValueError
        If `region` are not allowed ISO codes.
    """

    # If not Dataframe provided, download it
    if df is None:
        df = pd.read_csv(
            r"https://cnecovid.isciii.es/covid19/resources/casos_hosp_uci_def_sexo_edad_provres.csv"
        )
    # Select the columns of interest
    if response_var not in ["num_casos", "num_hosp", "num_uci", "num_def"]:
        raise ValueError("Provide a suitable response variable name.")
    df = df.filter(items=["provincia_iso", "sexo", "grupo_edad", "fecha", response_var])
    # Select the correct gender and region
    if gender is not None:
        if len(set(gender) - set(["H", "M"])) > 0:
            raise ValueError("Provide suitable gender categories.")
        df = df.loc[df["sexo"].isin(gender)]
    if region is not None:
        iso_codes = pd.read_csv("../data/provinces_es.csv")["code"]
        if len(set(region) - set(iso_codes)) > 0:
            raise ValueError("Provide suitable region codes.")
        df = df.loc[df["provincia_iso"].isin(region)]
    # Aggregate the response variable
    df = df.groupby(["grupo_edad", "fecha"]).agg({response_var: np.sum}).reset_index()
    # Convert fecha to datetime
    df["fecha"] = pd.to_datetime(df["fecha"], infer_datetime_format=True)
    # Make the group ages a categorical variable
    df["grupo_edad"] = df["grupo_edad"].astype("category")
    # If no minimum date is provided, take first non-zero response variable date
    if min_date is None:
        min_date = (df.groupby(["fecha"])[response_var].sum().cumsum() != 0).idxmax()
    mask_min_date = df["fecha"] >= min_date
    # If no maximum date is provided, take second maximum date available
    # (maximum date is set to 0 in the downloaded file)
    if max_date is None:
        max_date = df["fecha"].max()
    mask_max_date = df["fecha"] < max_date

    df = df[mask_max_date & mask_min_date]
    # Sort the dataframe by date and age group
    df = df.sort_values(by=["fecha", "grupo_edad"], ascending=[True, True])
    return df


def agg_covid_by_age(
    df: pd.DataFrame,
    response_var: str,
    normalize: bool = False,
    agg_method: Union[Callable, str] = np.sum,
) -> pd.DataFrame:

    """Aggregates the filtered COVID-19 DataFrame by age.

    Parameters
    ----------
    df : pd.DataFrame
        The filtered COVID-19 DataFrame.
    response_var : str
        The response variable chosen to filter the DataFrame.
    normalize : bool, optional
        If True, normalizes the data in the 0-1 range. By default, False.
    agg_method : Union[Callable, str], optional
        The aggregation method used. By default, np.sum.

    Returns
    -------
    pd.DataFrame
        The aggregated by age DataFrame.

    Raises
    ------
    ValueError
        If `response_var` is not an allowed response variable.
    """

    if response_var not in ["num_casos", "num_hosp", "num_uci", "num_def"]:
        raise ValueError("Provide a suitable response variable name.")
    df = df.groupby(["fecha"]).agg({response_var: agg_method})
    if normalize:
        df = normalize_data(df)
    return df


def pivot_covid_df(
    df: pd.DataFrame, response_var: str, normalize: bool = False
) -> pd.DataFrame:

    """
    Pivot the filtered COVID-19 DataFrame. The index of the final Dataframe are
    the dates and the columns correspond to the age group.

    df : pd.DataFrame
        The filtered COVID-19 DataFrame.
    response_var : str
        The response variable chosen to filter the DataFrame.
    normalize : bool, optional
        If True, normalizes the data in the 0-1 range. By default, False.

    Returns
    -------
    pd.DataFrame
        The pivoted COVID-19 DataFrame.

    Raises
    ------
    ValueError
        If `response_var` is not an allowed response variable.
    """

    if response_var not in ["num_casos", "num_hosp", "num_uci", "num_def"]:
        raise ValueError("Provide a suitable response variable name.")

    df = df.query("grupo_edad != 'NC'").pivot(
        index="fecha", columns="grupo_edad", values=response_var
    )
    if normalize:
        df = pd.DataFrame(
            data=normalize_data(df.values), index=df.index, columns=df.columns
        )
    return df


def get_days_from_covid_df(df: pd.DataFrame) -> np.ndarray:

    """
    Get the days from first date present in the filtered COVID-19 DataFrame.

    Parameters
    ----------
    df : pd.DataFrame
        The pivoted or the aggregated by age COVID-19 DataFrame.

    Returns
    -------
    np.ndarray
        The ordered array containing the days from first record.
    """

    return (
        (df.index.to_series().diff() / np.timedelta64(1, "D")).fillna(0).cumsum().values
    )


def displaced_forecast_covid(
    deriv: np.ndarray,
    xmax: Union[int, float],
    lag: Union[int, float] = 0,
    factor_deriv: Union[int, float] = 1,
) -> Dict[str, Union[np.ndarray, Union[int, float]]]:

    """
    Compute the most extreme point in the forecast prediction and the
    coordinates of the points where constraints are imposed.

    Parameters
    ----------
    deriv : np.ndarray
        Value of the growth rates in the first wave.
    xmax : Union[int, float]
        Last point of the fitting region.
    lag : Union[int, float], optional
        The lag factor of the points in the second wave. By default, 0.
    factor_deriv : Union[int, float], optional
        The derivative factor needed to multiplied the growth rates in the first
        wave. By default, 1.

    Returns
    -------
    Dict[str, Union[np.ndarray, Union[int, float]]]
        A dictionary containing the following items:
        - `x_last`: The most extreme point in the forecast prediction.
        - `x_pred`: The points where the growth rates are enforced.
        - `deriv_pred`: The prediction derivative values.
    """

    # Get the most extreme point in the forecast prediction
    x_last = xmax + lag * len(deriv)
    # Get the derivatives multiplied by the `factor_deriv`
    deriv_pred = np.array([elem * factor_deriv for elem in deriv])
    # Get the values where the derivatives of the curve are imposed
    x_pred = np.arange(xmax + lag, x_last + lag, lag)
    return {"x_last": x_last, "x_pred": x_pred, "deriv_pred": deriv_pred}



from typing import Dict, Iterable, Tuple, Union

import mosek.fusion
import numpy as np
from cpsplines.psplines.bspline_basis import BsplineBasis


class ObjectiveFunction:

    """
    Define the objective function of the optimization problem. The initial
    convex quadratically optimization problem is reformulated as an equivalent
    conic quadratic problem with linear objective function.

    Parameters
    ----------
    bspline : Iterable[BsplineBasis]
        An iterable containing the B-spline bases objects used to approximate
        the function to estimate.
    model : mosek.fusion.Model
        The MOSEK model of the problem.

    Attributes
    ----------
    var_dict: Dict[str, mosek.fusion.LinearVariable]
        The resulting variable dictionary from the method `_create_var_dict`.
    """

    def __init__(self, bspline: Iterable[BsplineBasis], model: mosek.fusion.Model):
        self.bspline = bspline
        self.model = model

    def _create_var_dict(self, n: int) -> Dict[str, mosek.fusion.LinearVariable]:

        """
        Creates the variables of the optimization problem. These variables are:

        - The coefficient multidimensional array ("theta") with dimensions
          k_1 x ... x k_N, where k_i is the cardinal of B-spline basis along the
          i-th dimension.
        - The artificial variable of the design matrix product ("t_B") included
          in the reformulation of the quadratic term.
        - The artificial variable of the i-th penalty matrix product ("t_D_i")
          when restating this summand as a rotated quadratic cone constraint.

        Parameters
        ----------
        n : int
            Number of smoothing parameters needed in the optimization function.

        Returns
        -------
        Dict[str, mosek.fusion.LinearVariable]
            A dictionary containing the variables involved on the objective
            function of the optimization problem. The keys of the dictionary are
            the names of the variables and the values are the variables
            themselves.
        """

        var_dict = {}
        # The coefficient multidimensional array
        variable_shape = [bsp.matrixB.shape[1] for bsp in self.bspline]
        var_dict["theta"] = self.model.variable(
            "theta",
            variable_shape,
            mosek.fusion.Domain.unbounded(),
        )
        for i in range(n):
            var_dict[f"t_D_{i}"] = self.model.variable(
                f"t_D_{i}", 1, mosek.fusion.Domain.greaterThan(0.0)
            )
        var_dict["t_B"] = self.model.variable(
            "t_B", 1, mosek.fusion.Domain.greaterThan(0.0)
        )
        return var_dict

    def create_obj_function(
        self,
        L_B: np.ndarray,
        L_D: Iterable[np.ndarray],
        sp: Iterable[Union[int, float]],
        lin_term: np.ndarray,
    ) -> Tuple[Union[None, mosek.fusion.ConicConstraint]]:

        """
        Creates the objective function to be minimized. Although the penalized
        sum of squares is directly stated as a quadratic optimization function,
        the positive semidefiniteness of the matrices in the quadratic term
        enables to reformulate it as a linear objective function together with
        rotated quadratic cone constraints.

        Indeed, each of the individual summands x^TAx in the quadratic term of
        the objective function may be replaced with a new artificial variable t
        and impose the constraint x^TAx <= t, which has an equivalent
        characterization as the (k+2)-vector (t, 1/2, L^Tx) belonging to the
        (k+2)-dimensional rotated quadratic cone, where L satisfies A = LL^T.

        Parameters
        ----------
        L_B : np.ndarray
            The lower triangular matrix from the Cholesky decomposition of B^TB,
            where B denotes the design matrix of the B-splines basis.
        L_D : Iterable[np.ndarray]
            An iterable containing the lower triangular matrices from the
            Cholesky decomposition of D^TD, where D denotes the difference
            matrix of the penalty term.
        sp : Iterable[Union[int, float]]
            An iterable containing the smoothing parameters.
        lin_term : np.ndarray
            An array containing the coefficients of the linear term in the
            objective function. This array is dot multiplied with the row major
            vectorization of the array constituted by the coefficients in the
            basis expansion of B-splines.

        References
        ----------
        - Alizadeh, F., & Goldfarb, D. (2003). Second-order cone programming.
          Mathematical programming, 95(1), 3-51.

        Returns
        -------
        Tuple[Union[None, mosek.fusion.ConicConstraint]]
            The linear objective function and the rotated quadratic cone
            constraints resulting from reformulating the summands on the
            quadratic term.

        Raises
        ------
        ValueError
            If lengths of the smoothing parameter vector and penalty matrix
            iterable differ.
        """

        if len(sp) != len(L_D):
            raise ValueError(
                "The number of smoothing parameters and penalty matrices must agree."
            )

        # Generate the decision variables involved in the objective function
        self.var_dict = self._create_var_dict(n=len(sp))

        # Linear term coefficients must be dot multiplied by the row major
        # vectorization of the basis expansion multidimensional array
        # coefficients
        flatten_theta = mosek.fusion.Var.flatten(self.var_dict["theta"])
        cons = []
        # Create the rotated quadratic cone constraint of B^TB
        cons.append(
            self.model.constraint(
                "rot_cone_B",
                mosek.fusion.Expr.vstack(
                    self.var_dict["t_B"],
                    1 / 2,
                    mosek.fusion.Expr.mul(
                        mosek.fusion.Matrix.sparse(L_B.T), flatten_theta
                    ),
                ),
                mosek.fusion.Domain.inRotatedQCone(),
            )
        )
        # Create the rotated quadratic cone constraint of each D^TD
        for i, L in enumerate(L_D):
            cons.append(
                self.model.constraint(
                    f"rot_cone_{i}",
                    mosek.fusion.Expr.vstack(
                        self.var_dict[f"t_D_{i}"],
                        1 / 2,
                        mosek.fusion.Expr.mul(
                            mosek.fusion.Matrix.sparse(L.T), flatten_theta
                        ),
                    ),
                    mosek.fusion.Domain.inRotatedQCone(),
                )
            )

        # The linear term from the original objective function
        obj = [mosek.fusion.Expr.dot(lin_term, flatten_theta)]
        # The rotated cone reformulation on the penalty term yield summands on
        # the objective function of the form sp*t_D, where t_D is the new
        # artificial variable introduced in the characterization
        for i, sp in enumerate(sp):
            obj.append(mosek.fusion.Expr.dot(sp, self.var_dict[f"t_D_{i}"]))
        # The rotated cone reformulation on the basis term yield a summand of
        # the artificial variable t_B included during the reformulation
        obj = mosek.fusion.Expr.add(self.var_dict["t_B"], mosek.fusion.Expr.add(obj))
        # Generate the minimization objective function object
        obj = self.model.objective(
            "obj",
            mosek.fusion.ObjectiveSense.Minimize,
            obj,
        )
        return tuple([obj] + cons)
